import datetime
import glob
import logging
import os
import re
from functools import partial

import h5py
import numpy as np
import vtk
from ccpi.viewer.utils import Converter
from ccpi.viewer.utils.conversion import (cilHDF5CroppedReader, cilHDF5ResampleReader, cilMetaImageCroppedReader,
                                          cilMetaImageResampleReader, cilNumpyCroppedReader, cilNumpyResampleReader,
                                          cilRawCroppedReader, cilRawResampleReader, cilTIFFCroppedReader,
                                          cilTIFFResampleReader)
from ccpi.viewer.utils.error_handling import EndObserver, ErrorObserver
from ccpi.viewer.utils.hdf5_io import HDF5Reader
#from ccpi.viewer.version import version
from schema import Optional, Or, Schema, SchemaError
from vtk.util import numpy_support
from vtk.util.vtkAlgorithm import VTKPythonAlgorithmBase


def SaveRenderToPNG(render_window, filename):
    ''' Saves contents of a vtk render window
    to a PNG file.

    Parameters
    ----------
    render_window: vtkRenderWindow
        render window to save contents of.
    filename: str
        name of file to write PNG to.
    '''
    w2if = vtk.vtkWindowToImageFilter()
    w2if.SetInput(render_window)
    w2if.Update()

    basename = os.path.splitext(os.path.basename(filename))[0]
    # Note regex matching is different to glob matching:
    regex = '{}_([0-9]*)\.png'.format(basename)
    fname_string = '{}_{}.png'.format(basename, '[0-9]*')
    directory = os.path.dirname(filename)
    slist = []

    for el in glob.glob(os.path.join(directory, fname_string)):
        el = os.path.basename(el)
        slist.append(int(re.search(regex, el).group(1)))

    if len(slist) == 0:
        number = 0
    else:
        number = max(slist) + 1

    saveFilename = '{}_{:04d}.png'.format(os.path.join(directory, basename), number)

    writer = vtk.vtkPNGWriter()
    writer.SetFileName(saveFilename)
    writer.SetInputConnection(w2if.GetOutputPort())
    writer.Write()


class ImageReader(object):
    '''
    Generic reader for reading to vtkImageData
    Currently reads: HDF5, MetaImage, Numpy, Raw, TIFF stacks
    Supports resampling OR cropping the dataset whilst
    reading.
    Currently doesn't support both cropping and resampling
    If set both to true then it resamples and doesn't crop
    '''

    def __init__(self,
                 file_name=None,
                 resample=True,
                 target_size=512**3,
                 crop=False,
                 target_z_extent=None,
                 resample_z=False,
                 raw_image_attrs=None,
                 hdf5_dataset_name="entry1/tomo_entry/data/data",
                 log_file=None):
        '''
        Constructor

        Parameters
        ----------
        file_name: os.path or string, default None
            file name to read
            In case of TIFF files, either the directory to the TIFF files, or the path to one such file is needed.
        resample: bool, default True
            whether to resample
        crop: bool, default False
            whether to crop
        target_size: int, default 512**3
            target size after downsampling
        target_z_extent: list [,], default None
            desired extent after cropping on z axis
        resample_z: bool, default True
            whether to resample on the z axis. E.g. in the case we have
            acquisition data, the projections would be on the z axis, so
            we would not want to resample in that direction.
        raw_image_attrs: dict, default None
            Attributes of the raw image data, must have the format:
            {'shape': 2D or 3D array, 'is_fortran':bool, 'is_big_endian':bool
            'type_code':info_var['typcode']}
        hdf5_dataset_name: string, default "entry1/tomo_entry/data/data"
            Name of the hdf5 dataset to be read, if file format is hdf5
        log_file: str, optional, default None
            log verbose output to file of this name            
        '''

        if file_name is None:
            raise Exception('Path to file is required.')

        if not (os.path.isfile(file_name) or os.path.isdir(file_name)):
            raise Exception('Path\n {}\n does not exist.'.format(file_name))

        self._OriginalImageAttrs = {}

        self.SetFileName(file_name)
        self.SetResample(resample)
        self.SetTargetSize(target_size)
        self.SetCrop(crop)
        self.SetTargetZExtent(target_z_extent)
        self.SetResampleZ(resample_z)
        self.SetHDF5DatasetName(hdf5_dataset_name)
        self.SetRawImageAttributes(raw_image_attrs)
        self.SetLogFileName(log_file)

    def SetFileName(self, file_name):
        '''
        Parameters
        ----------
        file_name: os.path or string, default None
            file name to read
        '''
        self._FileName = file_name

    def SetResample(self, resample):
        '''
        Parameters
        ----------
        resample: bool, default True
            whether to resample
        '''
        self._Resample = resample

    def SetTargetSize(self, target_size):
        '''
        Parameters
        ----------
        target_size: int, default 512**3
            target size after downsampling
        '''

        self._TargetSize = target_size

    def SetCrop(self, crop):
        '''
        Parameters
        ----------
        crop: bool, default False
            whether to crop
        '''
        self._Crop = crop

    def SetTargetZExtent(self, target_z_extent):
        '''
        Parameters
        ----------
        target_z_extent: list [,], default None
            desired extent after cropping on z axis
            '''
        self._TargetZExtent = target_z_extent

    def SetResampleZ(self, resample_z):
        '''
        Parameters
        ----------
        resample_z: bool, default True
            whether to resample on the z axis. E.g. in the case we have
            acquisition data, the projections would be on the z axis, so
            we would not want to resample in that direction.
        '''
        self._ResampleZ = resample_z

    def SetHDF5DatasetName(self, hdf5_dataset_name):
        '''
        Parameters
        ----------
        hdf5_dataset_name: string, default "entry1/tomo_entry/data/data"
            Name of the hdf5 dataset to be read, if file format is hdf5
        '''
        self._HDF5DatasetName = hdf5_dataset_name

    def SetLogFileName(self, log_file):
        '''
        Parameters
        ----------
        log_file: str, optional, default None
            log verbose output to file of this name            
        '''
        self._SetUpLogger(log_file)

    def SetRawImageAttributes(self, raw_image_attrs):
        if raw_image_attrs is not None and raw_image_attrs != {}:
            try:
                raw_attrs = self._ValidateRawAttrs(raw_image_attrs)
                self._OriginalImageAttrs = raw_attrs
            except SchemaError as e:
                raise ValueError("Error: Raw image attributes were not input correctly: ", e)

    def Read(self, *args, **kwargs):
        ''' reads self._FileName
            returns vtkImageData'''
        # identifies file type
        # uses appropriate reader based on file type and cropping or resampling

        if self._Crop:
            if self._TargetZExtent is None:
                raise TypeError("If crop is set to True, target_z_extent must be set.")
            if self._Resample:
                self.logger.warning(
                    "Both cropping and resampling is not yet implemented. Image will just be cropped and not resampled."
                )
                self._Resample = False

        self._LoadedImageAttrs = {'resampled': self._Resample, 'cropped': self._Crop}

        self.logger.info("reading: {}".format(self._FileName))

        progress_callback = kwargs.get('progress_callback')

        reader = self._GetReader(progress_callback)
        reader.Update()
        data = reader.GetOutput()

        self._UpdateLoadedImageAttrs(reader, data)

        self._UpdateOriginalImageAttrs(reader)

        return data

    def GetOriginalImageAttrs(self):
        return self._OriginalImageAttrs

    def GetLoadedImageAttrs(self):
        return self._LoadedImageAttrs

    def _ValidateRawAttrs(self, raw_image_attrs):
        if raw_image_attrs is None:
            return
        raw_attrs = raw_image_attrs.copy()
        raw_attrs_schema = Schema({
            'shape': Or(list, np.ndarray, tuple),
            'is_fortran': bool,
            'is_big_endian': bool,
            'typecode': str
        })
        raw_attrs_schema.validate(raw_attrs)
        return raw_attrs

    def _GetReader(self, progress_callback=None):
        '''
        Returns an appropriate reader for the image file provided.
        The appropriate reader is decided by matching the extension of the file to be read, which can be:
        - mha, mhd for METAIO files
        - npy for numpy file format
        - raw for binary blobs
        - nxs, h5, or hdf5 for HDF5 files
        - tiff or tif for TIFF stacks.
        No actual check of the file format is performed.
        '''
        if os.path.isfile(self._FileName):
            file_extension = os.path.splitext(self._FileName)[1]

            if file_extension in ['.mha', '.mhd']:
                reader = self._GetMetaImageReader()

            elif file_extension in ['.npy']:
                reader = self._GetNumpyImageReader()

            elif file_extension in ['.raw']:
                reader = self._GetRawImageReader()

            elif file_extension in ['.nxs', '.h5', '.hdf5']:
                reader = self._GetHDF5ImageReader()
                self._OriginalImageAttrs['dataset_name'] = self._HDF5DatasetName

            elif file_extension in ['.tif', '.tiff']:
                image_files = glob.glob(os.path.join(os.path.dirname(self._FileName), '*{}'.format(file_extension)))
                if len(image_files) == 0:
                    raise Exception('No tiff files were found in: {}'.format(os.path.dirname(self._FileName)))
                reader = self._GetTiffImageReader()
                reader.SetFileName(image_files)

            else:
                raise Exception('File format is not supported. Accepted formats include: .mhd, .mha, .npy, .tif, .raw')
        else:  # If we are given a folder, not a file, look for tiff files and try to read them
            image_files = list(
                glob.glob(os.path.join(self._FileName, '*.tif')) +
                list(glob.glob(os.path.join(self._FileName, '*.tiff'))))
            if len(image_files) == 0:
                raise Exception('No tiff files were found in: {}'.format(self._FileName))
            reader = self._GetTiffImageReader()
            reader.SetFileName(image_files)
            file_extension = '.tiff'

        if file_extension not in ['.tif', '.tiff']:
            # currently the tiff reader doesn't take these inputs:
            reader.SetFileName(self._FileName)
        else:
            image_files.sort(key=self.__natural_keys)

        # setting SetIsAcquisitionData determines whether to crop on Z:
        reader.SetIsAcquisitionData(not self._ResampleZ)

        if not self._Crop:
            if self._Resample:
                target_size = self._TargetSize
            else:
                # forced use of resample reader in the case that we
                # don't want to crop or resample,
                # but the large target size means we don't resample
                target_size = 1e12
            reader.SetTargetSize(int(target_size))

        # Add observers:
        reader.AddObserver(vtk.vtkCommand.ProgressEvent,
                           partial(self._ReportProgress, progress_callback=progress_callback))

        # Prints the error if an error occurs in the reader.
        # Otherwise this wouldn't print at all.
        error_obs = ErrorObserver(callback_fn=print)
        reader.AddObserver(vtk.vtkCommand.ErrorEvent, error_obs)
        # Could add end observer so that we don't continue to do anything
        # else if an error does occur in reader?

        return reader

    def __natural_keys(self, text):
        '''
        Used in the sorting of tiff files retrieved with glob
        https://stackoverflow.com/questions/5967500/how-to-correctly-sort-a-string-with-a-number-inside
        alist.sort(key=natural_keys) sorts in human order
        http://nedbatchelder.com/blog/200712/human_sorting.html
        (See Toothy's implementation in the comments)
        '''
        return [self.__atoi(c) for c in re.split(r'(\d+)', text)]

    def __atoi(self, text):
        '''Used in the sorting of tiff files retrieved with glob'''
        return int(text) if text.isdigit() else text

    def _GetMetaImageReader(self, progress_callback=None):
        if self._Crop:
            reader = cilMetaImageCroppedReader()
            reader.SetTargetZExtent(tuple(self._TargetZExtent))
        else:
            reader = cilMetaImageResampleReader()
        return reader

    def _GetNumpyImageReader(self, progress_callback=None):
        if self._Crop:
            reader = cilNumpyCroppedReader()
            reader.SetTargetZExtent(tuple(self._TargetZExtent))
        else:
            reader = cilNumpyResampleReader()

        return reader

    def _GetTiffImageReader(self, progress_callback=None):
        if self._Crop:
            reader = cilTIFFCroppedReader()
            reader.SetTargetZExtent(tuple(self._TargetZExtent))
        else:
            reader = cilTIFFResampleReader()
        return reader

    def _GetRawImageReader(self):
        if self._OriginalImageAttrs is None or 'shape' not in self._OriginalImageAttrs.keys():
            raise Exception("To read a raw image, raw_image_attrs must be set.")

        isFortran = self._OriginalImageAttrs['is_fortran']
        isBigEndian = self._OriginalImageAttrs['is_big_endian']
        typecode = self._OriginalImageAttrs['typecode']
        shape = tuple(self._OriginalImageAttrs['shape'])

        if self._Crop:
            reader = cilRawCroppedReader()
            reader.SetTargetZExtent(tuple(self._TargetZExtent))
        else:
            reader = cilRawResampleReader()

        reader.SetBigEndian(isBigEndian)
        reader.SetIsFortran(isFortran)
        reader.SetTypeCodeName(typecode)
        reader.SetStoredArrayShape(shape)

        return reader

    def _GetHDF5ImageReader(self):
        if self._Crop:
            reader = cilHDF5CroppedReader()
            reader.SetTargetExtent([0, -1, 0, -1, self._TargetZExtent[0], self._TargetZExtent[1]])

        else:
            reader = cilHDF5ResampleReader()

        reader.SetDatasetName(self._HDF5DatasetName)

        return reader

    def _SetUpLogger(self, fname=None, log_level=logging.INFO):
        """Set up the logger """
        self.logger = logging.getLogger("ccpi.viewer.utils.io.ImageReader")
        self.logger.setLevel(log_level)
        if fname is not None:
            handler = logging.FileHandler(fname)
            self.logger.addHandler(handler)

    def _ReportProgress(self, caller, event, progress_callback=None):
        ''' This emits the progress as a value between 1 and 100,
        and writes to a log file.
        If a Qt progress_callback has been passed, this allows progress to be kept track
        of if the reading is run in a Worker thread.'''
        progress_value = caller.GetProgress() * 100
        progress = "{:.1f}%".format(progress_value)
        self.logger.info(progress)

        if progress_callback is not None:
            progress_callback.emit(int(progress_value))

    def _UpdateLoadedImageAttrs(self, reader, data):
        # Make sure whether we did resample or not:
        if self._Resample:
            original_image_size = reader.GetStoredArrayShape()[0] * reader.GetStoredArrayShape(
            )[1] * reader.GetStoredArrayShape()[2]
            resampled_image_size = reader.GetTargetSize()
            if original_image_size <= resampled_image_size:
                self._LoadedImageAttrs['resampled'] = False
            else:
                self._LoadedImageAttrs['resampled'] = True
        # info about new dataset:
        self._LoadedImageAttrs['spacing'] = data.GetSpacing()
        self._LoadedImageAttrs['origin'] = data.GetOrigin()
        if self._Resample:
            self._LoadedImageAttrs['resample_z'] = self._ResampleZ

    def _UpdateOriginalImageAttrs(self, reader):
        self._OriginalImageAttrs['shape'] = reader.GetStoredArrayShape()
        self._OriginalImageAttrs['spacing'] = reader.GetElementSpacing()
        self._OriginalImageAttrs['origin'] = reader.GetOrigin()
        self._OriginalImageAttrs['bit_depth'] = str(reader.GetBytesPerElement() * 8)
        self._OriginalImageAttrs['is_big_endian'] = reader.GetBigEndian()
        self._OriginalImageAttrs['header_length'] = reader.GetFileHeaderLength()
        self._OriginalImageAttrs['file_name'] = self._FileName
        self._OriginalImageAttrs['resampled'] = False
        self._OriginalImageAttrs['cropped'] = False


class ImageWriterInterface(object):
    '''
    Interface for writers, with methods for setting and getting information about modified
    (i.e. resampled or cropped) image data to write to a file.
    
    The interface allows to set if the writer should use chunking for writing the file to disk.
    This is done by calling `SetChunking(True)` and by setting the chunk shape by using
    `SetChunkShape((10,10))`. By default no chunking is used.
    
    For the subclass `cilviewerHDF5Writer` the default chunking shape is set to one slice on the
    Z direction. 
    '''

    def __init__(self):
        self._FileName = None
        self._FileFormat = None
        self._OriginalDataset = None
        self._ChildDatasets = []
        self._OriginalDatasetAttributes = None
        self._ChildDatasetsAttributes = []
        self._Chunking = True
        self._ChunkShape = None
        self._HDF5Compression = None

    def SetFileName(self, value):
        '''
        Set the file name or path where to write the image data

        Parameters
        -----------
        value: (str)
            file name or path
        '''
        self._FileName = value

    def GetFileName(self):
        ''' The file name or path where to write the image data '''
        return self._FileName

    def SetFileFormat(self, value):
        '''
        Parameters
        -----------
        value: (str)
            file format
            must be one of: hdf5, nexus, mha
        '''
        self._FileFormat = value

    def GetFileFormat(self):
        return self._FileFormat

    def SetOriginalDataset(self, original_dataset, attributes):
        '''Parameters
        -----------
        original_dataset: vtkImageData or None
            original dataset, or None if this won't be saved to the file
        attributes: dict
            dictionary containing attributes of original dataset
        '''
        if not isinstance(original_dataset, vtk.vtkImageData) and not (original_dataset is None):
            raise Exception("'original_dataset' must be vtk.vtkImageData or None")

        self._validate_original_dataset_attributes(original_dataset, attributes)
        self._OriginalDataset = original_dataset
        self._OriginalDatasetAttributes = attributes

    def _validate_original_dataset_attributes(self, original_dataset, attributes):
        if not isinstance(attributes, dict):
            raise Exception("'attributes' must be a dictionary.")
        if original_dataset is None:
            # must have shape and filename set
            if attributes.get("file_name") is None or attributes.get("shape") is None:
                raise Exception(
                    "If no name is given for a dataset, the attributes must include the 'file_name' and the 'shape'.")

        original_attributes_schema = Schema({
            Optional('file_name'): str,
            Optional('shape'): Or(list, tuple),
            'resampled': False,
            'cropped': False,
            Optional(str): object  # allow any other keys and values
        })

        original_attributes_schema.validate(attributes)

    def AddChildDataset(self, child_dataset, attributes=None):
        if not isinstance(child_dataset, vtk.vtkImageData):
            raise Exception("child_dataset must be vtk.vtkImageData")
        # check type is vtkImageData
        self._ValidateChildDatasetAttributes(child_dataset, attributes)
        self._ChildDatasets.append(child_dataset)
        self._ChildDatasetsAttributes.append(attributes)

    def _ValidateChildDatasetAttributes(self, child_dataset, attributes):
        if not isinstance(attributes, dict) and not (attributes is None):
            raise Exception("'attributes' must be a dictionary, or unset (i.e. None)")

    def SetChunking(self, chunking):
        '''
        Parameters
        ----------
        chunking: bool, default: True in case of HDF5 file,
                  otherwise False
            Whether to write to the file in chunks.
        '''
        self._Chunking = chunking

    def SetChunkShape(self, chunk_shape):
        '''
        Parameters
        ----------
        chunk_shape: tuple, default: one slice on the z axis
            If SetChunking() has been set to True, this will be
            the size of a chunk written to the file.
        '''
        self._ChunkShape = chunk_shape

    def SetHDF5Compression(self, compression):
        '''
        Parameters
        ----------
        compression: list, default None
            This is a list of compression settings when writing to HDF5
            First element: str
                The type of hdf5 compression to use.
                Must be one of: None, 'gzip', or 'lzf'
            Second element:
                Options for the compression filter. For details, see:
                https://docs.h5py.org/en/stable/high/dataset.html?highlight=compression_opts#dataset-compression
            Third element: bool
                Whether to shuffle the data i.e. rearrange the bytes in
                a chunk, which may improve the compression ratio
        '''
        self._HDF5Compression = compression


class ImageWriter(ImageWriterInterface):
    '''
    Writer for writing out a modified i.e. resampled or cropped dataset.
    Currently supports writing to HDF5 and metaimage.
    
    In the case of HDF5:
    Expects to be writing an original dataset or attributes of the original dataset,
    plus one or more 'child' versions of the dataset which have been resampled and/or cropped.
    This can be done if the file format is set to hdf5 or nxs.
    If an image has been downsampled/cropped using the ImageReader class, then the attributes 
    obtained with: reader.get_original_image_attributes() and reader.get_loaded_image_attributes()
    will be in the correct format for this writer.
    
    Example of writing to HDF5:
    writer = ImageWriter()
    writer.SetFileName('resampled_image')
    writer.SetFileFormat('hdf5')
    writer.SetOriginalDataset(None, 
        {'file_name': 'image.nxs', 'shape': [500, 600, 600], 'resampled': False, 'cropped': False})
    writer.AddChildDataset(downsampled_image,  
        {'origin': [0.5, 0.5, 0.5], 'spacing': [2, 2, 2], 'resampled': True, 'resampled_z': False, 'cropped': False})
    writer.Write()

    Alternatively, to just write a downsampled/cropped dataset to a metaimage file with no extra attributes.
    Remember the metaimage writer already automatically will save the spacing and origin of the vtkImageData
    so there is no need to set this as extra attributes.
    Example of writing to MHA:
    writer = ImageWriter()
    writer.SetFileName('resampled_image')
    writer.SetFileFormat('mha')
    writer.AddChildDataset(downsampled_image)
    writer.Write()
    With mha this means we lose the information about the original dataset before it was downsampled/cropped

    '''

    def __init__(self):
        super(ImageWriter, self).__init__()

    def Write(self):
        # check file ext
        writer = self._GetWriter()
        writer.Write()

    def _GetWriter(self):
        file_name = os.path.splitext(self._FileName)[0]

        if self._FileFormat in ['nxs', 'h5', 'hdf5', '']:
            self._FileName = file_name + '.hdf5'
            writer = self._GetHDF5Writer()

        elif self._FileFormat in ['mha']:
            self._FileName = file_name + '.mha'
            writer = self._GetMetaImageWriter()

        else:
            raise Exception("File format is not supported. Supported types include hdf5/nexus.")

        writer.SetFileName(self._FileName)
        return writer

    def _GetHDF5Writer(self):
        writer = cilviewerHDF5Writer()
        writer.SetOriginalDataset(None, self._OriginalDatasetAttributes)
        writer.SetChunking(self._Chunking)
        writer.SetChunkShape(self._ChunkShape)
        writer.SetHDF5Compression(self._HDF5Compression)
        for i in range(0, len(self._ChildDatasets)):
            writer.AddChildDataset(self._ChildDatasets[i], self._ChildDatasetsAttributes[i])
        return writer

    def _GetMetaImageWriter(self):
        writer = vtk.vtkMetaImageWriter()
        writer.SetInputData(self._ChildDatasets[0])
        return writer


class cilviewerHDF5Writer(ImageWriterInterface):
    '''
    Expects to be writing an original dataset or attributes of the original dataset,
    plus one or more 'child' versions of the dataset which have been resampled and/or cropped.
    '''

    def __init__(self):
        super(cilviewerHDF5Writer, self).__init__()

    def _ValidateChildDatasetAttributes(self, child_dataset, attributes):
        if not isinstance(attributes, dict):
            raise Exception("'attributes must be a dictionary.")
        # check origin and spacing are set
        # check resampling and cropping are set
        child_attributes_schema = Schema({
            'resampled': bool,
            'cropped': bool,
            'spacing': Or(list, tuple),
            'origin': Or(list, tuple),
            Optional('resample_z'): bool,
            Optional(str): object  # allow any other keys and values
        })
        child_attributes_schema.validate(attributes)

    def Write(self):
        for var in [self._FileName, self._OriginalDatasetAttributes]:
            if var is None:
                raise Exception("file_name, dataset(/s) and attribute(/s), are required.")
        for var in [self._ChildDatasets, self._ChildDatasetsAttributes]:
            if var is []:
                raise Exception("child dataset(/s) and attribute(/s), are required.")

        with h5py.File(self._FileName, 'w') as f:

            # give the file some important attributes
            f.attrs['file_name'] = self._FileName
            #f.attrs['viewer_version'] = version
            f.attrs['file_time'] = str(datetime.datetime.utcnow())
            f.attrs['creator'] = np.string_('io.py')
            f.attrs['HDF5_Version'] = h5py.version.hdf5_version
            f.attrs['h5py_version'] = h5py.version.version

            # create the NXentry group
            nxentry = f.create_group('entry1/tomo_entry')
            nxentry.attrs['NX_class'] = 'NXentry'

            datasets = [self._OriginalDataset]
            datasets += self._ChildDatasets

            attributes = [self._OriginalDatasetAttributes]
            attributes += self._ChildDatasetsAttributes

            for i, dataset in enumerate(datasets):
                data = dataset
                dataset_info = attributes[i]
                entry_num = i + 1
                dataset_name = 'entry{}/tomo_entry/data/data'.format(entry_num)

                if data is not None:
                    # The function imgdata.GetPointData().GetScalars() returns a pointer to a
                    # vtk<TYPE>Array where the data is stored as X-Y-Z.
                    array = numpy_support.vtk_to_numpy(data.GetPointData().GetScalars())

                    # Note that we flip the shape here because
                    # VTK's order is Fortran whereas h5py writes in
                    # C order. We don't want to do deep copies so we write
                    # with shape flipped and pretend the array is
                    # C order.
                    array = array.reshape(data.GetDimensions()[::-1])
                else:
                    array = None
                try:
                    if array is None:
                        dset = f.create_dataset(dataset_name, dataset_info['shape'])
                    else:
                        if self._Chunking:
                            if self._ChunkShape is None:
                                # If Chunking has been selected but a shape has not,
                                # by default use a slice as the chunk
                                slice_shape = list(array.shape)
                                slice_shape[0] = 1
                                self._ChunkShape = tuple(slice_shape)
                            if self._HDF5Compression is None or self._HDF5Compression[0] is None:
                                dset = f.create_dataset(dataset_name, data=array, chunks=self._ChunkShape)
                            elif self._HDF5Compression[0] == 'gzip':
                                dset = f.create_dataset(dataset_name,
                                                        data=array,
                                                        chunks=self._ChunkShape,
                                                        compression=self._HDF5Compression[0],
                                                        compression_opts=self._HDF5Compression[1],
                                                        shuffle=self._HDF5Compression[2])
                            elif self._HDF5Compression[0] == 'lzf':
                                dset = f.create_dataset(dataset_name,
                                                        data=array,
                                                        chunks=self._ChunkShape,
                                                        compression=self._HDF5Compression[0],
                                                        shuffle=self._HDF5Compression[2])
                        else:
                            dset = f.create_dataset(dataset_name, data=array)

                except RuntimeError:
                    print("Unable to save image data to {0}."
                          "Dataset with name {1} already exists in this file.".format(self._FileName, dataset_name))

                if entry_num != 1:
                    dset.attrs['original_dataset'] = 'entry1/tomo_entry/data/data'

                for key, value in dataset_info.items():
                    dset.attrs[key] = value


class cilviewerHDF5Reader(HDF5Reader):
    '''
    Expects to be reading a file where:
    entry1 contains an original dataset or attributes of the original dataset
    The following entries contain:
    one or more 'child' versions of the dataset which have been resampled and/or cropped.
    It is one of these 'child' versions that we are interested in reading for displaying in the viewer.
    The user must specify which they would like if there is more than one. By default entry2 is read.
    '''

    def __init__(self):
        VTKPythonAlgorithmBase.__init__(self, nInputPorts=0, nOutputPorts=1, outputType='vtkImageData')

        super(cilviewerHDF5Reader, self).__init__()
        self._DatasetEntryNumber = 2
        self._DatasetName = 'entry{}/tomo_entry/data/data'.format(self._DatasetEntryNumber)

    def SetDatasetEntryNumber(self, num):
        '''
        The entry number that the dataset will be read from.
        By default this is set to 2 which means the dataset read is:
        'entry2/tomo_entry/data/data'
        entry numbers:
        1 - should contain the attributes of the unmodified dataset
        2 - should contain a downsampled/cropped version 
        3+ - if present, should contain another downsampled/cropped version 
        '''
        dataset_name = 'entry{}/tomo_entry/data/data'.format(num)
        if self._FileName is not None:
            with h5py.File(self._FileName, 'r') as f:
                if not (dataset_name in f):
                    raise Exception("No dataset named {} exists in {}.".format(dataset_name, self._FileName))
        self._DatasetEntryNumber = num
        self._DatasetName = dataset_name

    def GetDatasetEntryNumber(self):
        '''
        The entry number that the dataset will be read from.
        By default this is set to 2 which means the dataset read is:
        'entry2/tomo_entry/data/data'
        entry numbers:
        1 - should contain the attributes of the unmodified dataset
        2 - should contain a downsampled/cropped version 
        3+ - if present, should contain another downsampled/cropped version 
        '''
        return self._DatasetEntryNumber

    def SetDatasetName(self, lname):
        '''
        It is easier to use SetDatasetEntryNumber,
        but you may still set the name instead if you
        wish.
        '''
        super(cilviewerHDF5Reader, self).SetDatasetName(lname)
        re_str = '^entry([0-9]*)'
        try:
            self._DatasetEntryNumber = re.search(re_str, str).group(1)
        except AttributeError:
            # This means no match found so naming convention of dataset is
            # not as we expect, so we can't assign a dataset entry number.
            self._DatasetEntryNumber = None

    def RequestData(self, request, inInfo, outInfo):
        output = super(cilviewerHDF5Reader, self)._update_output_data(outInfo)
        with h5py.File(self._FileName, 'r') as f:
            attrs = f[self._DatasetName].attrs
            # TODO check on the errors if these attributes haven't been found:
            output.SetOrigin(attrs['origin'])
            output.SetSpacing(attrs['spacing'])
        return 1



from ccpi.viewer.utils.conversion import Converter
from ccpi.viewer.utils.conversion import vtkImageResampler
import unittest
import numpy as np


def calculate_target_downsample_shape(max_size, total_size, shape, acq=False):
    if not acq:
        xy_axes_magnification = np.power(max_size / total_size, 1 / 3)
        slice_per_chunk = int(1 / xy_axes_magnification)
    else:
        slice_per_chunk = 1
        xy_axes_magnification = np.power(max_size / total_size, 1 / 2)
    num_chunks = 1 + len([i for i in range(slice_per_chunk, shape[2], slice_per_chunk)])

    target_image_shape = (int(xy_axes_magnification * shape[0]), int(xy_axes_magnification * shape[1]), num_chunks)
    return target_image_shape


class TestVTKImageResampler(unittest.TestCase):

    def setUp(self):
        # Generate random 3D array and convert to VTK Image Data:
        np.random.seed(1)
        bits = 16
        self.bytes_per_element = int(bits / 8)
        self.input_3D_array = np.random.randint(10, size=(50, 10, 60), dtype=eval(f"np.uint{bits}"))
        self.input_vtk_image = Converter.numpy2vtkImage(self.input_3D_array)

    def test_vtk_resample_reader(self):
        # Tests image with correct target size is generated by resample reader:
        # Not a great test, but at least checks the resample reader runs
        # without crashing
        # TODO: improve this test
        reader = vtkImageResampler()
        reader.SetInputDataObject(self.input_vtk_image)
        target_size = 100
        reader.SetTargetSize(target_size)
        reader.Update()

        image = reader.GetOutput()
        extent = image.GetExtent()
        og_shape = np.shape(self.input_3D_array)
        resulting_shape = (extent[1] + 1, (extent[3] + 1), (extent[5] + 1))
        og_shape = (og_shape[2], og_shape[1], og_shape[0])
        og_size = og_shape[0] * og_shape[1] * og_shape[2] * self.bytes_per_element
        expected_shape = calculate_target_downsample_shape(target_size, og_size, og_shape)
        self.assertEqual(resulting_shape, expected_shape)

        # # Now test if we get the full image extent if our
        # # target size is larger than the size of the image:
        target_size = og_size * 2
        reader.SetTargetSize(target_size)
        reader.Update()
        image = reader.GetOutput()
        extent = image.GetExtent()
        expected_shape = og_shape
        resulting_shape = (extent[1] + 1, (extent[3] + 1), (extent[5] + 1))
        self.assertEqual(resulting_shape, expected_shape)
        resulting_array = Converter.vtk2numpy(image)
        np.testing.assert_array_equal(self.input_3D_array, resulting_array)

        # # Now test if we get the correct z extent if we set that we
        # # have acquisition data
        target_size = 100
        reader.SetTargetSize(target_size)
        reader.SetIsAcquisitionData(True)
        reader.Update()
        image = reader.GetOutput()
        extent = image.GetExtent()
        shape_not_acquisition = calculate_target_downsample_shape(target_size, og_size, og_shape, acq=True)
        expected_size = shape_not_acquisition[0] * \
            shape_not_acquisition[1]*shape_not_acquisition[2]
        resulting_shape = (extent[1] + 1, (extent[3] + 1), (extent[5] + 1))
        resulting_size = resulting_shape[0] * \
            resulting_shape[1]*resulting_shape[2]
        # angle (z direction) is first index in numpy array, and in cil
        # but it is the last in vtk.
        resulting_z_shape = extent[5] + 1
        og_z_shape = np.shape(self.input_3D_array)[0]
        self.assertEqual(resulting_size, expected_size)
        self.assertEqual(resulting_z_shape, og_z_shape)


if __name__ == '__main__':
    unittest.main()



import os
import os.path
import vtk
import sys

# this is the location where the cil-data module saves the datasets
data_dir = os.path.abspath(os.path.join(sys.prefix, 'share', 'cil'))


class DATA(object):

    @classmethod
    def dfile(cls):
        return None

    @classmethod
    def get(cls, **kwargs):
        ddir = kwargs.get('data_dir', data_dir)
        loader = TestData(data_dir=ddir)
        return loader.load(cls.dfile(), **kwargs)


class HEAD(DATA):

    @classmethod
    def dfile(cls):
        return TestData.HEAD


class TestData(object):
    '''Class to return test data
    
    provides 1 dataset:
    HEAD = 'head.mha'
    '''
    HEAD = 'head.mha'

    def __init__(self, **kwargs):
        self.data_dir = kwargs.get('data_dir', data_dir)

    def load(self, which, **kwargs):
        '''
        Return a test data of the requested image

        Parameters
        ----------
        which: str
           Image selector: HEAD

        Returns
        -------
        vtkImageData
            The loaded dataset
        '''
        if which not in [TestData.HEAD]:
            raise ValueError('Unknown TestData {}.'.format(which))

        data_file = os.path.join(self.data_dir, which)

        reader = vtk.vtkMetaImageReader()
        reader.SetFileName(data_file)
        reader.Update()
        return reader.GetOutput()



#
#   Copyright 2022 STFC, United Kingdom Research and Innovation
#
#   Author 2022 Samuel Jones
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
from trame.app import get_server

from ccpi.web_viewer.trame_viewer import TrameViewer

try:
    import matplotlib.pyplot as plt
except ImportError:
    # Add optional overload to allow plt.colormaps to be called without matplotlib
    from ccpi.viewer.utils import CILColorMaps

    class BackupColorMaps:

        @staticmethod
        def colormaps():
            return ['viridis', 'plasma', 'inferno', 'magma']

    plt = BackupColorMaps()

from trame.widgets import vuetify

from ccpi.viewer.CILViewer import CILViewer
from ccpi.viewer.CILViewer2D import SLICE_ORIENTATION_XY
from ccpi.web_viewer.camera_data import CameraData

server = get_server()
state, ctrl = server.state, server.controller

DEFAULT_SLICE = 32
INITIAL_IMAGE = "head.mha"


class TrameViewer3D(TrameViewer):

    def __init__(self, list_of_files=None):
        super().__init__(list_of_files=list_of_files, viewer=CILViewer)

        # Define attributes that will be constructed in methods outside of __init__

        # omin, omax = min, max for volume opacity - may be scalar or gradient range:
        self.omin = None
        self.omax = None
        # cmin, cmax = min, max for slice and volume colors - scalar range:
        self.cmin = None
        self.cmax = None
        self.windowing_defaults = None
        self.slice_window_range_defaults = None
        self.slice_level_default = None
        self.slice_window_default = None
        self.slice_interaction_col = None
        self.slice_interaction_row = None
        self.slice_interaction_section = None
        self.volume_interaction_col = None
        self.volume_interaction_row = None
        self.volume_interaction_section = None
        self.disable_2d = False
        self.disable_3d = False

        # Define UI elements in __init__ to quiet down Pep8
        self.model_choice = None
        self.background_choice = None
        self.toggle_slice_visibility = None
        self.slice_slider = None
        self.toggle_window_details_button = None
        self.orientation_radio_buttons = None
        self.auto_window_level_button = None
        self.slice_window_range_slider = None
        self.slice_window_slider = None
        self.slice_level_slider = None
        self.show_slice_histogram_switch = None
        self.toggle_volume_visibility = None
        self.opacity_radio_buttons = None
        self.color_choice = None
        self.clipping_switch = None
        self.color_slider = None
        self.windowing_range_slider = None
        self.reset_cam_button = None
        self.reset_defaults_button = None
        self.clipping_removal_button = None

        self.set_opacity_mapping("scalar")

        # Grab current pos and orientation for reset later.
        self.original_cam_data = CameraData(self.cil_viewer.ren.GetActiveCamera())

        self.update_slice_slider_data()

        self.create_drawer_ui_elements()

        self.construct_drawer_layout()

        self.layout.content.children = [
            vuetify.VContainer(fluid=True, classes="pa-0 fill-height", children=[self.html_view])
        ]

        # Setup default state
        self.set_default_button_state()

    def construct_drawer_layout(self):
        # The difference is that we use range slider instead of detailed sliders
        self.slice_interaction_col = vuetify.VCol([
            self.toggle_slice_visibility, self.slice_slider, self.orientation_radio_buttons,
            self.auto_window_level_button, self.show_slice_histogram_switch, self.toggle_window_details_button,
            self.slice_window_range_slider, self.slice_window_slider, self.slice_level_slider
        ])
        self.slice_interaction_row = vuetify.VRow(self.slice_interaction_col)
        self.slice_interaction_section = vuetify.VContainer(self.slice_interaction_row)

        self.volume_interaction_col = vuetify.VCol([
            self.toggle_volume_visibility, self.opacity_radio_buttons, self.color_choice, self.color_slider,
            self.clipping_switch, self.clipping_removal_button, self.windowing_range_slider
        ])
        self.volume_interaction_row = vuetify.VRow(self.volume_interaction_col)
        self.volume_interaction_section = vuetify.VContainer(self.volume_interaction_row)

        self.layout.drawer.children = [
            "Choose model to load", self.model_choice,
            vuetify.VDivider(), "Choose background color", self.background_choice,
            vuetify.VDivider(), self.slice_interaction_section,
            vuetify.VDivider(), self.volume_interaction_section,
            vuetify.VDivider(), self.reset_cam_button,
            vuetify.VDivider(), self.reset_defaults_button
        ]

    def create_drawer_ui_elements(self):
        # replace this with the list browser? # https://kitware.github.io/trame/docs/module-widgets.html#ListBrowser
        self.model_choice = self.create_model_selector()
        self.background_choice = self.create_background_selector()
        self.toggle_slice_visibility = self.create_toggle_slice_visibility()
        self.slice_slider = self.create_slice_slider(disabled=self.disable_2d)
        self.toggle_window_details_button = self.create_toggle_window_details_button(disabled=self.disable_2d)
        self.auto_window_level_button = self.create_auto_window_level_button(disabled=self.disable_2d)
        self.orientation_radio_buttons = self.create_orientation_radio_buttons(disabled=self.disable_2d)
        self.slice_window_range_slider = self.construct_slice_window_range_slider(disabled=self.disable_2d)
        self.slice_window_slider = self.construct_slice_window_slider(disabled=self.disable_2d)
        self.slice_level_slider = self.construct_slice_level_slider(disabled=self.disable_2d)
        self.show_slice_histogram_switch = self.construct_show_slice_histogram_switch(disabled=self.disable_2d)
        self.toggle_volume_visibility = self.create_toggle_volume_visibility()
        self.opacity_radio_buttons = self.create_opacity_radio_buttons()
        self.color_choice = self.create_color_choice_selector()
        self.clipping_switch = self.create_clipping_toggle()
        self.clipping_removal_button = self.create_clipping_removal_button()
        self.update_windowing_defaults()
        self.color_slider = self.construct_color_slider()
        self.windowing_range_slider = self.construct_windowing_slider()
        self.reset_cam_button = self.create_reset_camera_button()
        self.reset_defaults_button = self.create_reset_defaults_button()

    def create_reset_defaults_button(self):
        return vuetify.VBtn("Reset defaults", hide_details=True, dense=True, solo=True, click=self.reset_defaults)

    def construct_show_slice_histogram_switch(self, disabled: bool):
        return vuetify.VSwitch(
            label="Show slice histogram",
            v_model=("show_slice_histogram", False),
            hide_details=True,
            dense=True,
            solo=True,
            disabled=disabled,
        )

    def create_reset_camera_button(self):
        return vuetify.VBtn(
            "Reset Camera",
            hide_details=True,
            dense=True,
            solo=True,
            click=self.reset_cam,
        )

    def create_clipping_toggle(self):
        return vuetify.VSwitch(
            label="Toggle Clipping",
            v_model=("toggle_clipping", False),
            hide_details=True,
            dense=True,
            solo=True,
            disabled=self.disable_3d,
        )

    def create_clipping_removal_button(self):
        return vuetify.VBtn("Remove clipping",
                            hide_details=True,
                            dense=True,
                            solo=True,
                            disabled=self.disable_3d,
                            click=self.remove_clipping_plane)

    def create_color_choice_selector(self):
        return vuetify.VSelect(
            v_model=("color_map", "viridis"),
            items=("color_map_options", plt.colormaps()),
            hide_details=True,
            solo=True,
            disabled=self.disable_3d,
        )

    def create_opacity_radio_buttons(self):
        return vuetify.VRadioGroup(
            children=[
                vuetify.VRadio(ref="opacity_scalar_button", label="Scalar", value="scalar"),
                vuetify.VRadio(label="Gradient", value="gradient"),
            ],
            v_model=("opacity", "scalar"),
            label="Opacity mapping:",
            disabled=self.disable_3d,
        )

    def create_toggle_volume_visibility(self):
        return vuetify.VSwitch(
            label="3D Volume visibility",
            v_model=("volume_visibility", True),
            hide_details=True,
            dense=True,
            solo=True,
        )

    def create_toggle_slice_visibility(self):
        return vuetify.VSwitch(label="2D Slice visibility",
                               v_model=("slice_visibility", True),
                               hide_details=True,
                               dense=True,
                               solo=True)

    def construct_color_slider(self):
        if not self.window_level_sliders_are_percentages:
            # Use actual values
            min_value = self.cmin
            max_value = self.cmax
            step = 1
        else:
            # Use percentages
            min_value = 0
            max_value = 100
            step = 0.5

        v_model = ("coloring", self.coloring_defaults)

        return vuetify.VRangeSlider(label="Color range",
                                    hide_details=True,
                                    solo=True,
                                    v_model=v_model,
                                    min=min_value,
                                    max=max_value,
                                    step=step,
                                    disabled=self.disable_3d,
                                    thumb_label=True,
                                    style="max-width: 300px")

    def construct_windowing_slider(self):
        if not self.window_level_sliders_are_percentages:
            # Use actual values
            min_value = self.omin
            max_value = self.omax
            step = 1
        else:
            # Use percentages
            min_value = 0
            max_value = 100
            step = 0.5

        v_model = ("windowing", self.windowing_defaults)

        return vuetify.VRangeSlider(label="Windowing",
                                    hide_details=True,
                                    solo=True,
                                    v_model=v_model,
                                    min=min_value,
                                    max=max_value,
                                    step=step,
                                    disabled=self.disable_3d,
                                    thumb_label=True,
                                    style="max-width: 300px")

    def update_windowing_defaults(self, method="scalar"):
        self.update_slice_data()
        state["opacity"] = method
        # Set omin and omax after set_slice_defaults because set_slice only uses scalar whereas we need to support gradient in 3D
        # omin, max = min, max for volume opacity - may be scalar or gradient range
        self.omin, self.omax = self.cil_viewer.getImageMapRange((0., 100.), method)
        # cmin, cmax = min, max for slice and volume colors - scalar range
        self.cmin, self.cmax = self.cil_viewer.getImageMapRange((0., 100.), "scalar")

        if hasattr(self, 'window_level_sliders_are_percentages') and self.window_level_sliders_are_percentages:
            self.coloring_defaults = 5., 95.
            self.windowing_defaults = 80., 99.
        else:
            self.windowing_defaults = self.cil_viewer.getImageMapRange((80., 99.), method)
            # colors always set based on scalar mapping:
            self.coloring_defaults = self.cil_viewer.getImageMapRange((5., 95.), "scalar")


        if hasattr(self, "windowing_range_slider") and self.windowing_range_slider is not None \
                and hasattr(self, "color_slider") and self.color_slider is not None:
            self.windowing_range_slider = self.construct_windowing_slider()
            self.color_slider = self.construct_color_slider()
            self.slice_window_range_slider = self.construct_slice_window_range_slider(self.disable_2d)
            self.slice_level_slider = self.construct_slice_level_slider(self.disable_2d)
            self.slice_window_slider = self.construct_slice_window_slider(self.disable_2d)
            self.construct_drawer_layout()
            self.layout.flush_content()

        state["windowing"] = self.windowing_defaults
        state["coloring"] = self.coloring_defaults

    def load_file(self, file_name, windowing_method="scalar"):
        # Perform the load before updating the UI
        super().load_file(file_name, windowing_method=windowing_method)
        # Update default values, there is an assumption this will not be called in the __init__ of this class.
        self.update_slice_slider_data()
        self.update_windowing_defaults(windowing_method)
        self.original_cam_data = CameraData(self.cil_viewer.ren.GetActiveCamera())

        # Reset all the buttons and camera
        self.reset_defaults()

    def switch_render(self):
        self.cil_viewer.style.ToggleVolumeVisibility()
        self.cil_viewer.updatePipeline()
        if hasattr(self, 'html_view'):
            self.html_view.update()

    def switch_slice(self):
        if self.cil_viewer.imageSlice.GetVisibility():
            self.cil_viewer.imageSlice.VisibilityOff()
        else:
            self.cil_viewer.imageSlice.VisibilityOn()
        self.cil_viewer.updatePipeline()
        self.html_view.update()

    def change_color_map(self, cmap):
        self.cil_viewer.setVolumeColorMapName(cmap)
        self.cil_viewer.updatePipeline()
        self.html_view.update()

    def set_opacity_mapping(self, opacity):
        self.cil_viewer.setVolumeRenderOpacityMethod(opacity)
        self.update_windowing_defaults(opacity)
        self.cil_viewer.updateVolumePipeline()
        self.html_view.update()

    def change_windowing(self, min_value, max_value, windowing_method="scalar"):
        if self.window_level_sliders_are_percentages:
            if windowing_method == "scalar":
                self.cil_viewer.setScalarOpacityPercentiles(min_value, max_value)
            else:
                self.cil_viewer.setGradientOpacityPercentiles(min_value, max_value)
        else:
            if windowing_method == "scalar":
                self.cil_viewer.setScalarOpacityRange(min_value, max_value)
            else:
                self.cil_viewer.setGradientOpacityRange(min_value, max_value)
        if hasattr(self, "html_view"):
            self.html_view.update()

    def reset_cam(self):
        self.cil_viewer.adjustCamera(resetcamera=True)
        if hasattr(self, "original_cam_data"):
            self.original_cam_data.copy_data_to_other_camera(self.cil_viewer.ren.GetActiveCamera())
        if hasattr(self, "html_view"):
            self.html_view.update()

    def reset_defaults(self):
        self.set_default_button_state()
        self.reset_cam()

    def set_default_button_state(self):
        # Don't reset file name
        state["slice"] = self.default_slice
        state["orientation"] = f"{SLICE_ORIENTATION_XY}"
        state["opacity"] = "scalar"
        state["color_map"] = "viridis"
        # resets to window-level for slice based on 5th, 95th percentiles over volume:
        min, max = self.cil_viewer.getImageMapRange((5., 95.), "scalar")
        window, level = self.cil_viewer.getSliceWindowLevelFromRange(min, max)
        if hasattr(self, 'window_level_sliders_are_percentages') and self.window_level_sliders_are_percentages:
            state["windowing"] = self.windowing_defaults
            state["coloring"] = self.coloring_defaults
            # window level of slice:
            state["slice_window_percentiles"] = (5., 95.)
            state["slice_window_as_percentage"] = self.convert_value_to_percentage(window)
            state["slice_level_as_percentage"] = self.convert_value_to_percentage(level)
        else:
            state["windowing"] = self.windowing_defaults
            state["coloring"] = self.coloring_defaults
            # window level of slice:
            state["slice_window_range"] = (min, max)
            state["slice_window"] = window
            state["slice_level"] = level

        state["slice_visibility"] = True
        state["volume_visibility"] = True
        state["slice_detailed_sliders"] = False
        state["background_color"] = "cil_viewer_blue"
        state["toggle_clipping"] = False
        state["show_slice_histogram"] = False
        # Ensure 2D is on
        if not self.cil_viewer.imageSlice.GetVisibility():
            self.switch_slice()
        # Ensure 3D is on
        if (not self.cil_viewer.volume_render_initialised) or (not self.cil_viewer.volume.GetVisibility()):
            self.switch_render()
        # Reset clipping on the volume itself:
        if self.cil_viewer.volume_render_initialised:
            if self.cil_viewer.volume.GetMapper().GetClippingPlanes() is not None:
                self.cil_viewer.volume.GetMapper().RemoveAllClippingPlanes()
        if self.cil_viewer.clipping_plane_initialised:
            self.cil_viewer.style.SetVolumeClipping(False)
            self.remove_clipping_plane()

    def change_coloring(self, min_value, max_value):
        if self.window_level_sliders_are_percentages:
            self.cil_viewer.setVolumeColorPercentiles(min_value, max_value)
        else:
            self.cil_viewer.setVolumeColorRange(min_value, max_value)
        if hasattr(self, "html_view"):
            self.html_view.update()

    def change_window_level_detail_sliders(self, show_detailed):
        super().change_window_level_detail_sliders(show_detailed)

        # Reconstruct the detailed sliders
        self.slice_window_range_slider = self.construct_slice_window_range_slider(self.disable_2d)
        self.slice_window_slider = self.construct_slice_window_slider(self.disable_2d)
        self.slice_level_slider = self.construct_slice_level_slider(self.disable_2d)

        # Reconstruct the drawer and push it
        self.construct_drawer_layout()
        self.layout.flush_content()

    def change_slice_visibility(self, visibility):
        if visibility:
            self.cil_viewer.imageSlice.VisibilityOn()
            self.disable_2d = False
        else:
            self.cil_viewer.imageSlice.VisibilityOff()
            self.disable_2d = True
            state["show_slice_histogram"] = False
        self.create_drawer_ui_elements()
        self.layout.flush_content()
        self.cil_viewer.updatePipeline()

    def change_volume_visibility(self, visibility):
        if not self.cil_viewer.volume_render_initialised:
            self.cil_viewer.installVolumeRenderActorPipeline()

        if visibility:
            self.disable_3d = False
        else:
            self.disable_3d = True
        self.cil_viewer.style.SetVolumeVisibility(visibility)
        self.create_drawer_ui_elements()
        self.layout.flush_content()
        self.cil_viewer.updatePipeline()

    def change_clipping(self, clipping_on):
        if clipping_on:
            state["slice_visibility"] = False
        self.cil_viewer.style.SetVolumeClipping(clipping_on)
        self.cil_viewer.updatePipeline()

    def remove_clipping_plane(self):
        if hasattr(self.cil_viewer, "planew"):
            state["toggle_clipping"] = False
            self.cil_viewer.remove_clipping_plane()
            self.cil_viewer.getRenderer().Render()
            self.cil_viewer.updatePipeline()

    def show_slice_histogram(self, show_histogram):
        self.cil_viewer.updateSliceHistogram()
        if show_histogram:
            self.cil_viewer.histogramPlotActor.VisibilityOn()
        else:
            self.cil_viewer.histogramPlotActor.VisibilityOff()
        self.html_view.update()

    def change_slice_number(self, slice_number):
        self.cil_viewer.updateSliceHistogram()
        self.cil_viewer.setActiveSlice(slice_number)
        self.cil_viewer.updatePipeline()
        self.html_view.update()



import os
import unittest

import h5py
import numpy as np
import vtk
from ccpi.viewer.utils.conversion import Converter, cilHDF5ResampleReader, cilHDF5CroppedReader
from ccpi.viewer.utils.hdf5_io import (HDF5Reader, HDF5SubsetReader, write_image_data_to_hdf5)


def calculate_target_downsample_shape(max_size, total_size, shape, acq=False):
    if not acq:
        xy_axes_magnification = np.power(max_size / total_size, 1 / 3)
        slice_per_chunk = int(1 / xy_axes_magnification)
    else:
        slice_per_chunk = 1
        xy_axes_magnification = np.power(max_size / total_size, 1 / 2)
    num_chunks = 1 + len([i for i in range(slice_per_chunk, shape[2], slice_per_chunk)])

    target_image_shape = (int(xy_axes_magnification * shape[0]), int(xy_axes_magnification * shape[1]), num_chunks)
    return target_image_shape


class TestHDF5IO(unittest.TestCase):

    def setUp(self):
        # Generate random 3D array and write to HDF5:
        np.random.seed(1)
        self.input_3D_array = np.random.random(size=(5, 10, 6))
        self.hdf5_filename_3D = 'test_3D_data.h5'
        with h5py.File(self.hdf5_filename_3D, 'w') as f:
            f.create_dataset('ImageData', data=self.input_3D_array)

        # Generate random 4D array and write to HDF5:
        self.input_4D_array = np.random.random(size=(10, 7, 10, 6))
        self.hdf5_filename_4D = 'test_4D_data.h5'
        with h5py.File(self.hdf5_filename_4D, 'w') as f:
            f.create_dataset('ImageData', data=self.input_4D_array)

    def test_read_hdf5(self):
        # Write a numpy array to a HDF5 and then test using
        # HDF5Reader to read it back:
        reader = HDF5Reader()
        reader.SetFileName(self.hdf5_filename_3D)
        reader.SetDatasetName("ImageData")
        reader.Update()
        array_image_data = reader.GetOutputDataObject(0)
        read_array = Converter.vtk2numpy(array_image_data)
        np.testing.assert_array_equal(self.input_3D_array, read_array)

    def test_read_hdf5_channel(self):
        # Test reading a specific channel in a 4D dataset
        channel_index = 5
        reader = HDF5Reader()
        reader.SetFileName(self.hdf5_filename_4D)
        reader.SetDatasetName("ImageData")
        reader.Set4DSliceIndex(channel_index)
        reader.Set4DIndex(0)
        reader.Update()
        array_image_data = reader.GetOutputDataObject(0)
        read_array = Converter.vtk2numpy(array_image_data)
        np.testing.assert_array_equal(self.input_4D_array[channel_index], read_array)

    def test_hdf5_subset_reader(self):
        # With the subset reader: -----------------------------
        # Test cropping the extent of a dataset
        cropped_array = self.input_3D_array[1:3, 3:6, 0:3]
        reader = HDF5Reader()
        reader.SetFileName(self.hdf5_filename_3D)
        reader.SetDatasetName("ImageData")
        cropped_reader = HDF5SubsetReader()
        cropped_reader.SetInputConnection(reader.GetOutputPort())
        # NOTE: the extent is in vtk so is in fortran order, whereas
        # above we had the np array in C-order so x and y cropping swapped
        cropped_reader.SetUpdateExtent((0, 2, 3, 5, 1, 2))
        cropped_reader.Update()
        array_image_data = cropped_reader.GetOutputDataObject(0)
        read_cropped_array = Converter.vtk2numpy(array_image_data)
        np.testing.assert_array_equal(cropped_array, read_cropped_array)

    def test_read_cropped_hdf5_reader(self):
        # # With the Cropped reader: -----------------------------
        cropped_array = self.input_3D_array[1:3, 3:6, 0:3]
        reader = cilHDF5CroppedReader()
        reader.SetFileName(self.hdf5_filename_3D)
        reader.SetDatasetName("ImageData")
        reader.SetTargetExtent((0, 2, 3, 5, 1, 2))
        reader.Update()
        array_image_data = reader.GetOutput()
        read_cropped_array = Converter.vtk2numpy(array_image_data)
        np.testing.assert_array_equal(cropped_array, read_cropped_array)

    def test_read_cropped_hdf5_channel(self):
        # Test cropping the extent of a dataset
        channel_index = 5
        cropped_array = self.input_4D_array[1:2, channel_index, 3:6, 0:3]
        hdf5_filename = 'test_numpy_data.h5'
        with h5py.File(hdf5_filename, 'w') as f:
            f.create_dataset('ImageData', data=self.input_4D_array)
        reader = HDF5Reader()
        reader.SetFileName(self.hdf5_filename_4D)
        reader.SetDatasetName("ImageData")
        reader.Set4DSliceIndex(channel_index)
        reader.Set4DIndex(1)
        cropped_reader = HDF5SubsetReader()
        cropped_reader.SetInputConnection(reader.GetOutputPort())
        # NOTE: the extent is in vtk so is in fortran order, whereas
        # above we had the np array in C-order so x and y cropping swapped
        cropped_reader.SetUpdateExtent((0, 2, 3, 5, 1, 1))
        cropped_reader.Update()
        array_image_data = cropped_reader.GetOutputDataObject(0)
        read_cropped_array = Converter.vtk2numpy(array_image_data)
        np.testing.assert_array_equal(cropped_array, read_cropped_array)

    def test_write_hdf5(self):
        # Provides an example image with extent (-10, 10, -10, 10, -10, 10):
        image_source = vtk.vtkRTAnalyticSource()
        image_source.Update()
        image_data = image_source.GetOutput()
        numpy_data = Converter.vtk2numpy(image_data)

        self.hdf5_filename_RT = "test_image_data.hdf5"

        write_image_data_to_hdf5(self.hdf5_filename_RT, image_data, dataset_name='RTData')

        # Test reading hdf5:
        reader = HDF5Reader()
        reader.SetFileName(self.hdf5_filename_RT)
        reader.SetDatasetName('RTData')
        reader.Update()

        read_image_data = reader.GetOutputDataObject(0)
        read_numpy_data = Converter.vtk2numpy(read_image_data)
        np.testing.assert_array_equal(numpy_data, read_numpy_data)
        # currently fails because we don't save extent to hdf5:
        # self.assertEqual(image_data.GetExtent(), read_image_data.GetExtent())
        os.remove(self.hdf5_filename_RT)

    def test_hdf5_resample_reader(self):
        # Tests image with correct target size is generated by resample reader:
        # Not a great test, but at least checks the resample reader runs
        # without crashing
        # TODO: improve this test
        readerhdf5 = cilHDF5ResampleReader()
        readerhdf5.SetFileName(self.hdf5_filename_3D)
        readerhdf5.SetDatasetName("ImageData")
        target_size = 100
        readerhdf5.SetTargetSize(target_size)
        readerhdf5.Update()
        image = readerhdf5.GetOutput()
        extent = image.GetExtent()
        resulting_shape = (extent[1] + 1, (extent[3] + 1), (extent[5] + 1))
        og_shape = np.shape(self.input_3D_array)
        og_shape = (og_shape[2], og_shape[1], og_shape[0])
        og_size = og_shape[0] * og_shape[1] * og_shape[2] * readerhdf5.GetBytesPerElement()
        expected_shape = calculate_target_downsample_shape(target_size, og_size, og_shape)
        self.assertEqual(resulting_shape, expected_shape)

        # Now test if we get the full image extent if our
        # target size is larger than the size of the image:
        target_size = og_size * 2
        readerhdf5.SetTargetSize(target_size)
        readerhdf5.Update()
        image = readerhdf5.GetOutput()
        extent = image.GetExtent()
        expected_shape = og_shape
        resulting_shape = (extent[1] + 1, (extent[3] + 1), (extent[5] + 1))
        self.assertEqual(resulting_shape, expected_shape)
        resulting_array = Converter.vtk2numpy(image)
        np.testing.assert_array_equal(self.input_3D_array, resulting_array)

        # Now test if we get the correct z extent if we set that we
        # have acquisition data
        readerhdf5 = cilHDF5ResampleReader()
        readerhdf5.SetDatasetName("ImageData")
        readerhdf5.SetFileName(self.hdf5_filename_3D)
        target_size = 100
        readerhdf5.SetTargetSize(target_size)
        readerhdf5.SetIsAcquisitionData(True)
        readerhdf5.Update()
        image = readerhdf5.GetOutput()
        extent = image.GetExtent()
        shape_not_acquisition = calculate_target_downsample_shape(target_size, og_size, og_shape, acq=True)
        expected_size = shape_not_acquisition[0] * \
            shape_not_acquisition[1]*shape_not_acquisition[2]
        resulting_shape = (extent[1] + 1, (extent[3] + 1), (extent[5] + 1))
        resulting_size = resulting_shape[0] * \
            resulting_shape[1]*resulting_shape[2]
        # angle (z direction) is first index in numpy array, and in cil
        # but it is the last in vtk.
        resulting_z_shape = extent[5] + 1
        og_z_shape = np.shape(self.input_3D_array)[0]
        self.assertEqual(resulting_size, expected_size)
        self.assertEqual(resulting_z_shape, og_z_shape)

    def tearDown(self):
        files = [self.hdf5_filename_3D, self.hdf5_filename_4D]
        for f in files:
            os.remove(f)


if __name__ == '__main__':
    unittest.main()
